@@ -16,7 +16,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: server_backup.c,v 1.32.2.14 2003/10/14 15:37:52 priikone Exp $ */
+/* $Id: server_backup.c,v 1.32.2.15 2003/11/28 17:22:02 priikone Exp $ */
 
 #include "serverincludes.h"
 #include "server_internal.h"
@@ -688,6 +688,8 @@ void silc_server_backup_resume_router(SilcServer server,
       unsigned char data[4];
       SILC_LOG_DEBUG(("Backup resuming not allowed since we are still "
 		      "primary router"));
+      SILC_LOG_INFO(("Backup resuming not allowed since we are still "
+		     "primary router"));
       SILC_PUT32_MSB(SILC_SERVER_BACKUP_START, data);
       silc_server_packet_send(server, sock, SILC_PACKET_FAILURE, 0,
 			      data, 4, FALSE);
